# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Created by: Joshua Mason
# Created date: 11/03/2021
# Copyright 2021 InferStat Ltd

"""
This submodule includes facilities for operations such as converting positions to price predictions and vice versa.
"""


from copy import deepcopy
from typing import List, Tuple, Union
import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame

from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error

from sklearn.pipeline import make_pipeline, FeatureUnion
from sklearn.preprocessing import FunctionTransformer, Binarizer
from infertrade.utilities.performance import calculate_portfolio_performance_python
from infertrade.PandasEnum import PandasEnum, create_price_column_from_synonym


def pct_chg(x: Union[np.ndarray, pd.Series]) -> np.ndarray:
    """Percentage change between the current and a prior element.

    Args:
        x: A numpy.ndarray or pandas.Series object

    Returns:
        A numpy.ndarray with the results

    """
    x = x.astype("float64")

    if isinstance(x, pd.DataFrame):
        pc = x.pct_change().values.reshape(-1, 1)
    else:
        x = np.reshape(x, (-1,))
        x_df = pd.Series(x, name="x")
        pc = x_df.pct_change().values.reshape(-1, 1)

    return pc


def lag(x: Union[np.ndarray, pd.Series], shift: int = 1) -> np.ndarray:
    """Lag (shift) series by desired number of periods.

    Args:
        x: A numpy.ndarray or pandas.Series object
        shift: The number of periods by which to shift the input time series

    Returns:
        A numpy.ndarray with the results

    """
    x = x.astype("float64")
    lagged_array = np.roll(x, shift=shift, axis=0)
    lagged_array[:shift, :] = np.nan
    return lagged_array


def research_over_price_minus_one(x: Union[np.ndarray, pd.Series], shift: int) -> np.ndarray:
    """Difference of two lagged log series.

    Args:
        x: A numpy.ndarray or pandas.Series object with exactly two columns
        shift: The number of periods by which the lag both series

    Returns:
        A numpy.ndarray with the results

    """
    x = np.array(x)
    x = x.astype("float64")

    if x.shape[1] != 2:
        raise IndexError(f"Number of columns must be 2.")

    lag_trans = FunctionTransformer(lag, kw_args={"shift": shift})

    pmr_pipe = make_pipeline(lag_trans)
    lagged = pmr_pipe.fit_transform(x)

    pmr = lagged[:, [1]] / lagged[:, [0]] - 1
    return pmr


class PricePredictionFromSignalRegression(TransformerMixin, BaseEstimator):

    """This class creates price predictions from signal values.

    Attributes:
        market_to_trade: The name of the column which contains the historical prices.

    """

    def __init__(self, market_to_trade: str = None):
        """Construction method for class PricePredictionFromPositions.

        Args:
            market_to_trade: The name of the column which contains the historical prices.

        Returns:
            None

        """
        if not market_to_trade:
            # We default to "price" as the target.
            market_to_trade = PandasEnum.MID.value
        self.market_to_trade = market_to_trade

    def fit(self, X: np.array, y=None):
        self.fitted_features_and_target_ = None
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """This method transforms a signal input to a price prediction.

        Args:
            X: A pandas.DataFrame object

        Returns:
            A pandas.DataFrame object

        """
        X_ = deepcopy(X)

        create_price_column_from_synonym(X_)

        regression_period = 120
        forecast_period = min(regression_period, len(X_))
        prediction_indices = self._get_model_prediction_indices(len(X_), regression_period, forecast_period)
        self._fit_features_matrix_target_array(X_)
        historical_signal_levels, historical_price_moves = self._get_features_matrix_target_array(X_)

        for ii_day in range(len(prediction_indices)):
            model_idx = prediction_indices[ii_day]["model_idx"]
            prediction_idx = prediction_indices[ii_day]["prediction_idx"]

            # Fit model
            regression_period_signal = historical_signal_levels[model_idx, :]
            regression_period_price_change = historical_price_moves[model_idx]

            rolling_regression_model = LinearRegression().fit(regression_period_signal, regression_period_price_change)
            # Predictions
            current_research = historical_signal_levels[prediction_idx, :]
            forecast = rolling_regression_model.predict(current_research)

            # Apply the calculated allocation to the dataframe.
            X_.loc[prediction_idx, PandasEnum.FORECAST_PRICE_CHANGE.value] = forecast

        if len(prediction_indices) == 0:
            X_[PandasEnum.FORECAST_PRICE_CHANGE.value] = 0
        else:
            X_[PandasEnum.FORECAST_PRICE_CHANGE.value].shift(-1)
        return X_

    def _get_features_matrix_transformer(self) -> ColumnTransformer:
        """
        1. Percent change of research series as predictor.
        2. Research series level as predictor.
        """
        percent_change_trans = FunctionTransformer(pct_chg)
        lag_1 = FunctionTransformer(lag, kw_args={"shift": 1})

        lag_pct = make_pipeline(lag_1, percent_change_trans)

        lp_m_lr_l1 = FunctionTransformer(research_over_price_minus_one, kw_args={"shift": 1})

        features = ColumnTransformer(
            [
                ("signal", lag_1, ["signal"]),
                ("signal_changes", lag_pct, ["signal"]),
                ("signal_differences", lp_m_lr_l1, [self.market_to_trade, "signal"]),
            ]
        )
        self.feature_names = ["signal", "signal_changes", "signal_differences"]
        return features

    def _get_features_matrix_target_array(self, input_time_series: pd.DataFrame) -> Tuple[pd.Series, pd.Series]:
        """Returns the target array features."""
        feat_tar_arr = self.fitted_features_and_target_.transform(input_time_series)
        feat_tar_arr = np.nan_to_num(feat_tar_arr, nan=0.0, posinf=0.0, neginf=0.0)

        features = np.delete(feat_tar_arr, -1, axis=1)
        target = feat_tar_arr[:, -1]

        return features, target

    def _fit_features_matrix_target_array(self, X: pd.DataFrame):
        """Get features matrix and target array. TODO -  more description helpful."""
        features = self._get_features_matrix_transformer()
        target = self._get_target_array_transformer()
        feat_tar = FeatureUnion(transformer_list=[("features", features), ("target", target)])
        self.fitted_features_and_target_ = feat_tar.fit(X)

    def _get_target_array_transformer(self):
        """Use level of price series as target (dependant) variable."""
        percent_change_trans = FunctionTransformer(pct_chg)
        target = ColumnTransformer([("historical_price_moves", percent_change_trans, [self.market_to_trade])])
        self.target_name = ["historical_price_moves"]
        return target

    @staticmethod
    def _get_model_prediction_indices(series_length: int, reg_period: int, forecast_period: int) -> List[dict]:
        """
        Create list of ranges for rolling regression.

        Parameters
        ----------
        series_length - total length of series
        reg_period - regression period
        forecast_period - forecast period

        Returns
        -------
        - model_idx are ranges for model fitting
        - prediction_idx are ranges for forecasting

        Examples
        --------
        {'model_idx': range(0, 50), 'prediction_idx': range(50, 60)}
        {'model_idx': range(10, 60), 'prediction_idx': range(60, 70)}
        {'model_idx': range(20, 70), 'prediction_idx': range(70, 80)}
        {'model_idx': range(30, 80), 'prediction_idx': range(80, 90)}
        {'model_idx': range(40, 90), 'prediction_idx': range(90, 100)}
        """

        indices_for_prediction = []

        for i in range(0, series_length - reg_period, forecast_period):
            # idx for model
            ind_start = i
            ind_end = i + reg_period

            # indices_for_prediction
            ind_pred_start = ind_end
            ind_pred_end = ind_pred_start + forecast_period
            if ind_pred_end > series_length:
                ind_pred_end = series_length

            indices_for_prediction.append(
                {"model_idx": range(ind_start, ind_end), "prediction_idx": range(ind_pred_start, ind_pred_end)}
            )

        return indices_for_prediction


class PositionsFromPricePrediction(TransformerMixin, BaseEstimator):

    """This class calculates the positions to take assuming Kelly Criterion."""

    def __init__(self):
        """Construction method for class PositionsFromPricePrediction.

        Args:
            None

        Returns:
            None

        """
        pass

    def fit(self, X, y=None):
        """This method is not used."""
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """This method calculates the positions to be taken based on the forecast price, assuming the Kelly Criterion.

        Args:
            X: A pandas.DataFrame object

        Returns:
            A pandas.DataFrame object

        """
        X_ = deepcopy(X)
        volatility = 0.1
        kelly_fraction = 1.0
        kelly_recommended_optimum = X[PandasEnum.FORECAST_PRICE_CHANGE.value] / volatility ** 2
        rule_recommended_allocation = kelly_fraction * kelly_recommended_optimum
        X_[PandasEnum.ALLOCATION.value] = rule_recommended_allocation
        return X_


class PricePredictionFromPositions(TransformerMixin, BaseEstimator):
    """This class converts positions into implicit price predictions based on the Kelly Criterion and an assumed volatility."""

    def __init__(self):
        """Construction method for class PricePredictionFromPositions.

        Args:
            None

        Returns:
            None

        """

        pass

    def fit(self, X, y=None):
        """This method is not used."""
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """This method converts allocations into the forecast one-day price changes.

        Args:
            X: A pandas.DataFrame object

        Returns:
            A pandas.DataFrame object

        """

        X_ = deepcopy(X)
        volatility = 0.1
        kelly_fraction = 1.0

        kelly_recommended_optimum = X_[PandasEnum.ALLOCATION.value] / kelly_fraction
        X_["PandasEnum.FORECAST_PRICE_CHANGE.value"] = kelly_recommended_optimum * volatility ** 2
        return X_


class ReturnsFromPositions(TransformerMixin, BaseEstimator):
    """This class calculates returns from positions."""

    def __init__(self):
        """Construction method for class ReturnsFromPositions.

        Args:
            None

        Returns:
            None

        """
        pass

    def fit(self, X, y=None):
        """This method is not used."""
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """This method converts positions into the cumulative portfolio return.

        Args:
            X: A pandas.DataFrame object

        Returns:
            A pandas.DataFrame object

        """
        X_1 = deepcopy(X)
        X_2 = deepcopy(X)
        X_1[PandasEnum.VALUATION.value] = calculate_portfolio_performance_python(X_2)[PandasEnum.VALUATION.value]
        return X_1


def limit_allocation(
    dataframe: pd.DataFrame, allocation_lower_limit: Union[int, float], allocation_upper_limit: Union[int, float]
) -> pd.DataFrame:
    """
    This function limits the ranges by limiting upper and lower limit of allocated values

    params:
    allocated_dataframe
    allocation_lower_limit: the lower limit for allocation values.
    allocation_upper_limit: the upper limit for allocation values.

    returns:
    allocation limited dataframe
    """
    if allocation_lower_limit > allocation_upper_limit:
        raise ValueError(
            "The lower limit for allocation values should not be greater than the upper limit for" " allocation values."
        )
    dataframe.loc[
        dataframe[PandasEnum.ALLOCATION.value] > allocation_upper_limit, PandasEnum.ALLOCATION.value
    ] = allocation_upper_limit
    dataframe.loc[
        dataframe[PandasEnum.ALLOCATION.value] < allocation_lower_limit, PandasEnum.ALLOCATION.value
    ] = allocation_lower_limit
    return dataframe


def daily_stop_loss(dataframe: pd.DataFrame, loss_limit: float) -> pd.DataFrame:
    """
    This function calculates loss and limit the allocation accordingly.
    It restricts allocation to 0 if loss>loss_limit

    params:
    allocated_dataframe
    loss_limit

    returns:
    dataframe
    """

    prev_alloc = 0
    prev_price = 0
    stop_loss_has_triggered = False
    for index, row in dataframe.iterrows():
        price_change = row.price - prev_price
        loss = -price_change * prev_alloc
        if loss > loss_limit or stop_loss_has_triggered:
            stop_loss_has_triggered = True
            row.allocation = 0
        prev_alloc = row.allocation
        prev_price = row.price
    return dataframe


def restrict_allocation(allocation_function: callable) -> callable:
    """
    This function is intended to be used as a decorator that may apply one or more restrictions to functions that calculate
    allocation values.
    """

    def restricted_function(*args, **kwargs) -> pd.DataFrame:

        # Get allocation dataframe from allocation function
        dataframe = allocation_function(*args, **kwargs)

        # Restriction by limiting allocations between range

        # initialize limits
        allocation_lower_limit = 0
        allocation_upper_limit = 1

        if "allocation_lower_limit" in kwargs:
            allocation_lower_limit = kwargs.get("allocation_lower_limit")
            dataframe = limit_allocation(dataframe, allocation_lower_limit, allocation_upper_limit)

        if "allocation_lower_limit" in kwargs:
            allocation_upper_limit = kwargs.get("allocation_upper_limit")
            dataframe = limit_allocation(dataframe, allocation_lower_limit, allocation_upper_limit)

        # Restriction by daily stop loss
        loss_limit = None
        if "loss_limit" in kwargs:
            loss_limit = kwargs.get("loss_limit")
            dataframe = daily_stop_loss(dataframe, loss_limit)

        return dataframe

    return restricted_function


def calculate_regression_with_kelly_optimum(
    df: pd.DataFrame,
    feature_matrix: pd.Series,
    last_feature_row: np.ndarray,
    target_array: pd.Series,
    regression_period: int,
    forecast_period: int,
    kelly_fraction: float = 1.0,
):

    # Refactor to make original method static.
    dataframe = df.copy()
    prediction_indices = PricePredictionFromSignalRegression._get_model_prediction_indices(
        series_length=len(feature_matrix), reg_period=regression_period, forecast_period=forecast_period
    )

    if not len(prediction_indices) > 0:
        raise IndexError("Unexpected error: Prediction indices are zero in length.")

    for ii_day in range(len(prediction_indices)):
        model_idx = prediction_indices[ii_day]["model_idx"]
        prediction_idx = prediction_indices[ii_day]["prediction_idx"]
        regression_period_signal = feature_matrix[model_idx, :]
        regression_period_price_change = target_array[model_idx]

        std_price = np.std(regression_period_price_change)
        std_signal = np.std(regression_period_signal)

        if not std_price > 0.0 or not std_signal > 0.0:
            if not std_price > 0.0:
                print("WARNING - price had no variation: ", std_price)
            if not std_signal > 0.0:
                print(
                    "WARNING - signal had no variation. Usually this means the lookback period was too short"
                    " for the data sample: ",
                    std_signal,
                )
            rule_recommended_allocation = 0.0
            volatility = 1.0
        else:
            # Assuming no bad inputs we calculate the recommended allocation.
            rolling_regression_model = LinearRegression().fit(regression_period_signal, regression_period_price_change)

            # Calculate model error
            predictions = rolling_regression_model.predict(regression_period_signal)
            forecast_horizon_model_error = np.sqrt(mean_squared_error(regression_period_price_change, predictions))

            # Predictions
            forecast_distance = 1
            current_research = feature_matrix[prediction_idx, :]
            forecast_price_change = rolling_regression_model.predict(current_research)

            # Calculate drift and volatility
            volatility = ((1 + forecast_horizon_model_error) * (forecast_distance ** -0.5)) - 1

            # Kelly recommended optimum
            if volatility < 0:
                raise ZeroDivisionError("Volatility needs to be positive value.")
            if volatility == 0:
                volatility = 0.01

            kelly_recommended_optimum = forecast_price_change / volatility ** 2
            rule_recommended_allocation = kelly_fraction * kelly_recommended_optimum

        # Apply the calculated allocation to the dataframe.
        dataframe.loc[prediction_idx, PandasEnum.ALLOCATION.value] = rule_recommended_allocation.reshape(
            -1,
        )

    # Shift position series  (QUESTION - does not appear to shift?)
    dataframe[PandasEnum.ALLOCATION.value] = dataframe[PandasEnum.ALLOCATION.value].shift(-1)

    # Calculate price forecast for last research value
    if std_price > 0.0 and std_signal > 0.0:
        # last_research = [[dataframe[PandasEnum.SIGNAL.value].iloc[-1]]]
        last_research = last_feature_row
        last_forecast_price = rolling_regression_model.predict(last_research)[0]
        value_to_update = kelly_fraction * (last_forecast_price / volatility ** 2)
    else:
        value_to_update = 0.0
    dataframe.iloc[-1, dataframe.columns.get_loc(PandasEnum.ALLOCATION.value)] = value_to_update

    return dataframe


def scikit_allocation_factory(allocation_function: callable) -> FunctionTransformer:
    """This function creates a SciKit Learn compatible Transformer embedding the position calculation.

    Args:
        allocation_function: A function to be turned into a sklearn.preprocessing.FunctionTransformer

    Returns:
        A sklearn.preprocessing.FunctionTransformer

    """
    return FunctionTransformer(allocation_function)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Created by: Thomas Oliver
# Created date: 17th March 2021
# Copyright 2021 InferStat Ltd

"""
PandasEnum for providing the special string names used with the InferTrade interface.
"""

from enum import Enum

import pandas as pd


class PandasEnum(Enum):

    """
    Provides the strings for special column names that InferTrade uses in identifying pandas dataframe contents.

    These strings should not be used for other purposes.

    Meanings:
    MID - this is the mid price used to calculate performance.

    ALLOCATION - the fraction of the overall portfolio the strategy wants to invest in the market. May differ from the
     amount invested where the strategy requires minimum deviations to trigger position adjustment.

    VALUATION - the value of strategy, after running a hypothetical rule implementing the strategy. 1.0 = 100% means no
     profit or loss. 0.9 = 90%, means a -10% cumulative loss. 1.1 = 110% means a 10% overall cumulative gain.

    BID_OFFER_SPREAD - the fractional bid-offer spread - 2 * (ask - bid)/(ask + bid) - for that time period.

    SIGNAL - an information time series that could be used for construction of an allocation series.
    """

    # Core string labels
    MID = "price"
    SIGNAL = "signal"
    FORECAST_PRICE_CHANGE = "forecast_price_change"
    ALLOCATION = "allocation"
    VALUATION = "portfolio_return"
    BID_OFFER_SPREAD = "bid_offer_spread"

    # Price synonyms
    CLOSE = "close"  # alternative name for mid at period end.
    ADJUSTED_CLOSE = "adjusted close"
    ADJ_CLOSE = "adj close"
    ADJ_DOT_CLOSE = "adj. close"
    PRICE_SYNONYMS = [CLOSE, ADJUSTED_CLOSE, ADJ_CLOSE, ADJ_DOT_CLOSE]

    # Diagnostic string labels
    PERIOD_START_CASH = "period_start_cash"
    PERIOD_START_SECURITIES = "period_start_securities"
    PERIOD_START_ALLOCATION = "start_of_period_allocation"
    PERIOD_END_CASH = "period_end_cash"
    PERIOD_END_SECURITIES = "period_end_securities"
    PERIOD_END_ALLOCATION = "end_of_period_allocation"
    PERCENTAGE_TO_BUY = "trade_percentage"
    TRADING_SKIPPED = "trading_skipped"
    SECURITIES_BOUGHT = "security_purchases"
    CASH_INCREASE = "cash_flow"


def create_price_column_from_synonym(df_potentially_missing_price_column: pd.DataFrame):
    """If the price column is missing then we will look for the "close" instead and copy those values."""
    if PandasEnum.MID.value not in df_potentially_missing_price_column.columns:
        successfully_updated = False
        for ii_synonym in PandasEnum.PRICE_SYNONYMS.value:
            if ii_synonym in df_potentially_missing_price_column.columns:
                df_potentially_missing_price_column[PandasEnum.MID.value] = df_potentially_missing_price_column[
                    ii_synonym
                ]
                successfully_updated = True
                break

        if not successfully_updated:
            raise KeyError(
                "Price column is missing - cannot perform transform. Needs one of these present: ",
                PandasEnum.MID.value,
                PandasEnum.CLOSE.value,
            )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2021 InferStat Ltd

"""
Functions for signals and positions created within this package.
"""
from copy import deepcopy

from infertrade.PandasEnum import PandasEnum
from infertrade.algos.community.allocations import infertrade_export_allocations
from infertrade.algos.community.signals import normalised_close, scikit_signal_factory, infertrade_export_signals
from infertrade.algos.community.ta_regressions import ta_export_regression_allocations

joint_set = deepcopy(infertrade_export_allocations)
joint_set.update(ta_export_regression_allocations)

# A dictionary providing the list of community signals and trading strategies.
infertrade_export = {
    PandasEnum.SIGNAL.value: infertrade_export_signals,
    PandasEnum.ALLOCATION.value: joint_set,
}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Nikola Rokvic
# Created: 17.08.2021
# Copyright 2021 InferStat Ltd

from copy import deepcopy

import pandas as pd
from sklearn.pipeline import make_pipeline
from infertrade.algos.community import scikit_signal_factory
from infertrade.algos import algorithm_functions, ta_adaptor
from infertrade.utilities.operations import PricePredictionFromSignalRegression, PositionsFromPricePrediction
from infertrade.utilities.performance import calculate_portfolio_performance_python
from infertrade.algos.community.ta_regressions import ta_export_regression_allocations
from infertrade.api import Api


def calculate_ta_regression_allocation(dataframe: pd.DataFrame, rule_name: str):
    """Used to implement ta regression allocations with market data."""
    df = deepcopy(dataframe)
    percent_gain = [0]

    ta_alloc = ta_export_regression_allocations[rule_name]["function"]
    df_with_allocations = ta_alloc(df)
    df_with_performance = calculate_portfolio_performance_python(df_with_allocations, detailed_output=True)
    for i in range(1, len(df_with_performance["portfolio_return"])):
        percent_gain.append(
            (df_with_performance["portfolio_return"][i] - df_with_performance["portfolio_return"][i - 1]) * 100
        )
    df_with_performance = df_with_performance.assign(percent_gain=percent_gain)
    return df_with_performance


def calculate_infertrade_allocation(dataframe: pd.DataFrame, rule_name: str):
    """Used to implement infertrade allocations with market data."""
    df = deepcopy(dataframe)
    percent_gain = [0]

    df_with_allocations = Api.calculate_allocations(df=df, name_of_strategy=rule_name, name_of_price_series="close")
    df_with_performance = calculate_portfolio_performance_python(df_with_allocations, detailed_output=True)
    for i in range(1, len(df_with_performance["portfolio_return"])):
        percent_gain.append(
            (df_with_performance["portfolio_return"][i] - df_with_performance["portfolio_return"][i - 1]) * 100
        )
    df_with_performance = df_with_performance.assign(percent_gain=percent_gain)
    return df_with_performance


def calculate_infertrade_signal(dataframe: pd.DataFrame, rule_name: str):
    """Used to implement infertrade signals with market data."""
    df = deepcopy(dataframe)
    percent_gain = [0]

    pipeline = make_pipeline(
        scikit_signal_factory(algorithm_functions["infertrade"]["signal"][rule_name]["function"]),
        PricePredictionFromSignalRegression(),
        PositionsFromPricePrediction(),
    )

    df = pipeline.fit_transform(df)
    df_with_performance = calculate_portfolio_performance_python(df, detailed_output=True)
    for i in range(1, len(df_with_performance["portfolio_return"])):
        percent_gain.append(
            (df_with_performance["portfolio_return"][i] - df_with_performance["portfolio_return"][i - 1]) * 100
        )
    df_with_performance = df_with_performance.assign(percent_gain=percent_gain)
    return df_with_performance


def calculate_ta_allocation(dataframe: pd.DataFrame, rule_name: str):
    "Used to implement ta allocations with market data"
    df = deepcopy(dataframe)
    percent_gain = [0]

    df_with_allocations = Api.calculate_allocations(df=df, name_of_strategy=rule_name, name_of_price_series="close")
    df_with_performance = calculate_portfolio_performance_python(df_with_allocations, detailed_output=True)
    for i in range(1, len(df_with_performance["portfolio_return"])):
        percent_gain.append(
            (df_with_performance["portfolio_return"][i] - df_with_performance["portfolio_return"][i - 1]) * 100
        )
    df_with_performance = df_with_performance.assign(percent_gain=percent_gain)
    return df_with_performance


def calculate_ta_signal(dataframe: pd.DataFrame, rule_name: str):
    "Used to implement ta signals with market data"
    df = deepcopy(dataframe)
    percent_gain = [0]

    adapted_rule = ta_adaptor(algorithm_functions["ta"]["signal"][rule_name]["class"], rule_name)

    pipeline = make_pipeline(
        scikit_signal_factory(adapted_rule),
        PricePredictionFromSignalRegression(),
        PositionsFromPricePrediction(),
    )
    df = pipeline.fit_transform(df)
    df_with_performance = calculate_portfolio_performance_python(df, detailed_output=True)
    for i in range(1, len(df_with_performance["portfolio_return"])):
        percent_gain.append(
            (df_with_performance["portfolio_return"][i] - df_with_performance["portfolio_return"][i - 1]) * 100
        )
    df_with_performance = df_with_performance.assign(percent_gain=percent_gain)
    return df_with_performance


def export_performance_df(
    dataframe: pd.DataFrame, rule_name: str, second_df: pd.DataFrame = None, relationship: str = None
) -> pd.DataFrame:
    """
    Function used to calculate portfolio performance for data after calculating a trading signal/rule and relationship.
    """
    if rule_name in algorithm_functions["infertrade"]["allocation"].keys():
        used_calculation = calculate_infertrade_allocation

    elif rule_name in algorithm_functions["ta"]["signal"].keys():
        used_calculation = calculate_ta_signal

    elif rule_name in algorithm_functions["ta"]["allocation"].keys():
        used_calculation = calculate_ta_allocation

    elif rule_name in algorithm_functions["infertrade"]["signal"].keys():
        used_calculation = calculate_infertrade_signal

    elif rule_name in ta_export_regression_allocations.keys():
        used_calculation = calculate_ta_regression_allocation

    elif rule_name not in algorithm_functions:
        raise ValueError("Algorithm not found")

    df_with_performance = used_calculation(dataframe=dataframe, rule_name=rule_name)
    if relationship is not None:
        if second_df is not None:
            second_df_with_performance = used_calculation(dataframe=second_df, rule_name=rule_name)
            second_df_with_relationship = calculate_infertrade_allocation(
                dataframe=second_df_with_performance, rule_name=relationship
            )

            df_with_relationship = calculate_infertrade_allocation(
                dataframe=df_with_performance, rule_name=relationship
            )

            complete_relationship = df_with_relationship.append(second_df_with_relationship, ignore_index=False)
            return complete_relationship
        else:
            df_with_relationship = calculate_infertrade_allocation(
                dataframe=df_with_performance, rule_name=relationship
            )
            return df_with_relationship
    else:
        return df_with_performance



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2021 InferStat Ltd
# Created by: Joshua Mason
# Created date: 11/03/2021

"""
Functions used to compute signals. Signals may be used for visual inspection or as inputs to trading rules.
"""

import pandas as pd
import numpy as np
from pandas.core.frame import DataFrame
from sklearn.preprocessing import FunctionTransformer
from ta.trend import (
    adx,
    adx_neg,
    adx_pos,
    macd_signal,
    sma_indicator,
    wma_indicator,
    ema_indicator,
    dpo,
    trix,
    stc,
    aroon_up,
    aroon_down,
    vortex_indicator_neg,
    vortex_indicator_pos,
)
from ta.momentum import ppo_signal, rsi, stochrsi, pvo_signal, tsi, kama, roc
from ta.volatility import (
    AverageTrueRange,
    UlcerIndex,
    bollinger_hband,
    bollinger_lband,
    bollinger_mavg,
    ulcer_index as ulcerindex,
)
from infertrade.algos.external.ta import ta_adaptor
from infertrade.PandasEnum import PandasEnum


def normalised_close(df: pd.DataFrame) -> pd.DataFrame:
    """
    Scales the close by the maximum value of the close across the whole price history.

    Note that this signal cannot be determined until the end of the historical period and so is unlikely to be suitable
     as an input feature for a trading strategy.
    """
    df["signal"] = df["close"] / max(df["close"])
    return df


def high_low_diff(df: pd.DataFrame) -> pd.DataFrame:
    """Calculates the range between low and high price."""
    df["signal"] = df["high"] - max(df["low"])
    return df


def high_low_diff_scaled(df: pd.DataFrame, amplitude: float = 1) -> pd.DataFrame:
    """Example signal based on high-low range times scalar."""
    df["signal"] = (df["high"] - max(df["low"])) * amplitude
    return df


def simple_moving_average(df: pd.DataFrame, window: int = 1) -> pd.DataFrame:
    """
    Calculates smooth signal based on price trends by filtering out the noise from random short-term price fluctuations
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = sma_indicator(df_with_signal["close"], window=window)
    return df_with_signal


def weighted_moving_average(df: pd.DataFrame, window: int = 1) -> pd.DataFrame:
    """
    Weighted moving averages assign a heavier weighting to more current data points since they are more relevant than data points in the distant past.
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = wma_indicator(df_with_signal["close"], window=window)
    return df_with_signal


def exponentially_weighted_moving_average(df: pd.DataFrame, window: int = 50) -> pd.DataFrame:
    """
    This function uses an exponentially weighted multiplier to give more weight to recent prices.
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = ema_indicator(df["close"], window=window, fillna=True)
    return df_with_signal


def moving_average_convergence_divergence(
    df: pd.DataFrame, window_slow: int = 26, window_fast: int = 12, window_signal: int = 9
) -> pd.DataFrame:
    """
    This function is a trend-following momentum indicator that shows the relationship between two moving averages at different windows:
    The MACD is usually calculated by subtracting the 26-period exponential moving average (EMA) from the 12-period EMA.

    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = macd_signal(df["close"], window_slow, window_fast, window_signal, fillna=True)
    return df_with_signal


def relative_strength_index(df: pd.DataFrame, window: int = 14) -> pd.DataFrame:
    """
    This function measures the magnitude of recent price changes to evaluate overbought or oversold conditions in the price.
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = rsi(df["close"], window=window, fillna=True)
    return df_with_signal


def stochastic_relative_strength_index(df: pd.DataFrame, window: int = 14) -> pd.DataFrame:
    """
    This function applies the Stochastic oscillator formula to a set of relative strength index (RSI) values rather
    than to standard price data.
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = stochrsi(df["close"], window=window, fillna=True)
    return df_with_signal


def chande_kroll(
    df: pd.DataFrame,
    average_true_range_periods: int = 10,
    average_true_range_multiplier: float = 1.0,
    stop_periods: int = 9,
) -> pd.DataFrame:
    """
    Calculates signals for the Chande-Kroll stop.

    See here: https://www.tradingview.com/support/solutions/43000589105-chande-kroll-stop
    """

    # Calculate the maximum and minum prices that the asset attained in the last
    # average_true_range_periods periods
    max_in_n_periods = df["high"].rolling(window=average_true_range_periods).max()
    min_in_n_periods = df["low"].rolling(window=average_true_range_periods).min()

    # Calculate the Average True Range indicator using average_true_range_periods periods, and
    # temporarily store those values in df[PandasEnum.SIGNAL.value]
    adapted_average_true_range = ta_adaptor(
        AverageTrueRange, "average_true_range", window=average_true_range_periods, fillna=False
    )

    signal_transformer = scikit_signal_factory(adapted_average_true_range)
    signal_transformer.fit_transform(df)

    # Calculate the intermediate high and low stops
    intermediate_high_stops = max_in_n_periods - (df[PandasEnum.SIGNAL.value] * average_true_range_multiplier)
    intermediate_low_stops = min_in_n_periods + (df[PandasEnum.SIGNAL.value] * average_true_range_multiplier)

    # Obtain the stop long and stop short values
    df["chande_kroll_long"] = intermediate_high_stops.rolling(window=stop_periods).max()
    df["chande_kroll_short"] = intermediate_low_stops.rolling(window=stop_periods).min()

    # Drop the Average True Range indicator values before returning
    df.drop(columns=PandasEnum.SIGNAL.value, inplace=True)

    return df


def bollinger_band(df: pd.DataFrame, window: int = 20, window_dev: int = 2) -> pd.DataFrame:
    """
    This function calculates signal which characterizes the prices and volatility over time.
    There are three lines that compose Bollinger Bands. By default:
        1. Middle band: A 20 day simple moving average
        2. The upper band: 2 standard deviations above from a 20 day simple moving average
        3. The lower band: 2 standard deviations - from a 20 day SMA

    These can be adjusted by changing parameter window and window_dev

    Parameters:
    window: Smoothing period
    window_dev: number of standard deviation
    """
    df_with_signal = df.copy()
    df_with_signal["typical_price"] = (df["close"] + df["low"] + df["high"]) / 3
    df_with_signal["BOLU"] = bollinger_hband(df_with_signal["typical_price"], window=window, window_dev=window_dev)
    df_with_signal["BOLD"] = bollinger_lband(df_with_signal["typical_price"], window=window, window_dev=window_dev)
    df_with_signal["BOLA"] = bollinger_mavg(df_with_signal["typical_price"], window=window)

    return df_with_signal


def detrended_price_oscillator(df: pd.DataFrame, window: int = 20) -> pd.DataFrame:
    """
    This function is a detrended price oscillator which strips out price trends in an effort to
    estimate the length of price cycles from peak to peak or trough to trough.
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = dpo(df["close"], window=window)
    return df_with_signal


def percentage_price_oscillator(
    df: pd.DataFrame, window_slow: int = 26, window_fast: int = 12, window_signal: int = 9
) -> pd.DataFrame:
    """
    This is a technical momentum indicator that shows the relationship between two moving averages in percentage terms.
    The moving averages are a 26-period and 12-period exponential moving average (EMA).
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = ppo_signal(df["close"], window_slow, window_fast, window_signal, fillna=True)
    return df_with_signal


def percentage_volume_oscillator(
    df: pd.DataFrame, window_slow: int = 26, window_fast: int = 12, window_signal: int = 9
) -> pd.DataFrame:
    """
    This is a technical momentum indicator that shows the relationship between two moving averages of volume in percentage terms.
    The moving averages are a 26-period and 12-period exponential moving average (EMA) of volume.
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = pvo_signal(df["volume"], window_slow, window_fast, window_signal, fillna=True)
    return df_with_signal


def triple_exponential_average(df: pd.DataFrame, window: int = 14) -> pd.DataFrame:
    """
    The triple exponential average (TRIX) is a momentum indicator shows the percentage change
    in a moving average that has been smoothed exponentially three times.
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = trix(df["close"], window, fillna=True)
    return df_with_signal


def true_strength_index(
    df: pd.DataFrame, window_slow: int = 25, window_fast: int = 13, window_signal: int = 13
) -> pd.DataFrame:
    """
    This is a technical momentum oscillator that finds trends and reversals.
    It helps in determining overbought and oversold conditions.
    It also gives warning of trend weakness through divergence.
    """
    df_with_signal = df.copy()
    df_with_signal["TSI"] = tsi(df["close"], window_slow, window_fast, fillna=True)
    df_with_signal["signal"] = ema_indicator(df_with_signal["TSI"], window_signal, fillna=True)
    return df_with_signal


def schaff_trend_cycle(
    df: pd.DataFrame, window_slow: int = 50, window_fast: int = 23, cycle: int = 10, smooth1: int = 3, smooth2: int = 3
) -> pd.DataFrame:
    """
    The Schaff Trend Cycle (STC) is a trend indicator that
    is commonly used to identify market trends and provide buy
    and sell signals to traders.

    Assumption:
    Currency trends accelerate and decelerate in cyclical patterns regardless of time frame
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = stc(
        df_with_signal["close"], window_slow, window_fast, cycle, smooth1, smooth2, fillna=True
    )
    return df_with_signal


def KAMA(df: pd.DataFrame, window: int = 10, pow1: int = 2, pow2: int = 30) -> pd.DataFrame:
    """
    Kaufman's Adaptive Moving Average (KAMA) is an indicator that
    indicates both the volatility and trend of the market.
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = kama(df["close"], window, pow1, pow2)
    return df_with_signal


def aroon(df: pd.DataFrame, window: int = 25) -> pd.DataFrame:
    """
    The Arron indicator is composed of two lines.
        1. Aroon_up: line which measures the number of periods since a High, and
        2. Aroon_down: line which measures the number of periods since a Low.
    """
    df_with_signal = df.copy()
    df_with_signal["aroon_up"] = aroon_up(df["close"], window, fillna=True)
    df_with_signal["aroon_down"] = aroon_down(df["close"], window, fillna=True)
    return df_with_signal


def rate_of_change(df: pd.DataFrame, window: int = 12) -> pd.DataFrame:
    """
    Rate of Change is momentum-based technical indicator that measures the percentage change in price between the current price and the price a certain number of periods ago.
    """
    df_with_signal = df.copy()
    df_with_signal["signal"] = roc(df["close"], window, fillna=True)
    return df_with_signal


def average_directional_movement_index(df: pd.DataFrame, window: int = 14) -> pd.DataFrame:
    """
    Average Directional Movement Index makes use of three indicators to measure both trend direction and its strength.
        1. Plus Directional Indicator (+DI)
        2. Negative Directonal Indicator (-DI)
        3. Average directional Index (ADX)
    +DI and -DI measures the trend direction and ADX measures the strength of trend
    """
    df_with_signal = df.copy()
    df_with_signal["ADX_POS"] = adx_pos(df["high"], df["low"], df["close"], window, fillna=True)
    df_with_signal["ADX_NEG"] = adx_neg(df["high"], df["low"], df["close"], window, fillna=True)
    df_with_signal["ADX"] = adx(df["high"], df["low"], df["close"], window, fillna=True)
    return df_with_signal


def vortex_indicator(df: pd.DataFrame, window: int = 14) -> pd.DataFrame:
    """
    A vortex indicator is used to detect trend reversals and confirm current trends.
    It is composed of two lines:
        1. an uptrend line (VI+) and
        2. a downtrend line (VI-)
    """
    df_with_signal = df.copy()
    df_with_signal["VORTEX_POS"] = vortex_indicator_pos(df["high"], df["low"], df["close"], window, fillna=True)
    df_with_signal["VORTEX_NEG"] = vortex_indicator_neg(df["high"], df["low"], df["close"], window, fillna=True)

    return df_with_signal


# creates wrapper classes to fit sci-kit learn interface
def scikit_signal_factory(signal_function: callable):
    """A class compatible with Sci-Kit Learn containing the signal function."""
    return FunctionTransformer(signal_function)


infertrade_export_signals = {
    "normalised_close": {
        "function": normalised_close,
        "parameters": {},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/e49334559ac5707db0b2261bd47cd73504a68557/infertrade/algos/community/signals.py#L31"
        },
    },
    "high_low_diff": {
        "function": high_low_diff,
        "parameters": {},
        "series": ["high", "low"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/e49334559ac5707db0b2261bd47cd73504a68557/infertrade/algos/community/signals.py#L42"
        },
    },
    "high_low_diff_scaled": {
        "function": high_low_diff_scaled,
        "parameters": {"amplitude": 1.0},
        "series": ["high", "low"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/e49334559ac5707db0b2261bd47cd73504a68557/infertrade/algos/community/signals.py#L153"
        },
    },
    "simple_moving_average": {
        "function": simple_moving_average,
        "parameters": {"window": 1},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/e49334559ac5707db0b2261bd47cd73504a68557/infertrade/algos/community/signals.py#L158"
        },
    },
    "weighted_moving_average": {
        "function": weighted_moving_average,
        "parameters": {"window": 1},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/e49334559ac5707db0b2261bd47cd73504a68557/infertrade/algos/community/signals.py#L60"
        },
    },
    "exponentially_weighted_moving_average": {
        "function": exponentially_weighted_moving_average,
        "parameters": {"window": 50},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/e49334559ac5707db0b2261bd47cd73504a68557/infertrade/algos/community/signals.py#L69"
        },
    },
    "moving_average_convergence_divergence": {
        "function": moving_average_convergence_divergence,
        "parameters": {"window_slow": 26, "window_fast": 12, "window_signal": 9},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L76"
        },
    },
    "relative_strength_index": {
        "function": relative_strength_index,
        "parameters": {"window": 14},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L100"
        },
    },
    "stochastic_relative_strength_index": {
        "function": stochastic_relative_strength_index,
        "parameters": {"window": 14},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L114"
        },
    },
    "chande_kroll": {
        "function": chande_kroll,
        "parameters": {"average_true_range_periods": 10, "average_true_range_multiplier": 1.0, "stop_periods": 9},
        "series": ["high", "low"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/e49334559ac5707db0b2261bd47cd73504a68557/infertrade/algos/community/signals.py#L125"
        },
    },
    "bollinger_band": {
        "function": bollinger_band,
        "parameters": {"window": 20, "window_dev": 2},
        "series": ["close", "high", "low"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5f74bdeb99eb26c15df0b5417de837466cefaee1/infertrade/algos/community/signals.py#L155"
        },
    },
    "detrended_price_oscillator": {
        "function": detrended_price_oscillator,
        "parameters": {"window": 20},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5f74bdeb99eb26c15df0b5417de837466cefaee1/infertrade/algos/community/signals.py#L186"
        },
    },
    "percentage_price_oscillator": {
        "function": percentage_price_oscillator,
        "parameters": {"window_slow": 26, "window_fast": 12, "window_signal": 9},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L196"
        },
    },
    "percentage_volume_oscillator": {
        "function": percentage_volume_oscillator,
        "parameters": {"window_slow": 26, "window_fast": 12, "window_signal": 9},
        "series": ["volume"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L204"
        },
    },
    "triple_exponential_average": {
        "function": triple_exponential_average,
        "parameters": {"window": 14},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L215"
        },
    },
    "true_strength_index": {
        "function": true_strength_index,
        "parameters": {"window_slow": 25, "window_fast": 13},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L228"
        },
    },
    "schaff_trend_cycle": {
        "function": schaff_trend_cycle,
        "parameters": {"window_slow": 50, "window_fast": 23, "cycle": 10, "smooth1": 3, "smooth2": 3},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L240"
        },
    },
    "KAMA": {
        "function": KAMA,
        "parameters": {"window": 10, "pow1": 2, "pow2": 30},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L257"
        },
    },
    "aroon": {
        "function": aroon,
        "parameters": {"window": 25},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L268"
        },
    },
    "rate_of_change": {
        "function": rate_of_change,
        "parameters": {"window": 12},
        "series": ["close"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L285"
        },
    },
    "rate_of_change": {
        "function": rate_of_change,
        "parameters": {"window": 12},
        "series": ["close", "high", "low"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L297"
        },
    },
    "average_directional_movement_index": {
        "function": average_directional_movement_index,
        "parameters": {"window": 14},
        "series": ["close", "high", "low"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L306"
        },
    },
    "vortex_indicator": {
        "function": vortex_indicator,
        "parameters": {"window": 14},
        "series": ["close", "high", "low"],
        "available_representation_types": {
            "github_permalink": "https://github.com/ta-oliver/infertrade/blob/5aa01970fc4277774bd14f0823043b4657e3a57f/infertrade/algos/community/signals.py#L321"
        },
    },
}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2021 InferStat Ltd
# Created by: Joshua Mason
# Created date: 11/03/2021

"""
Allocation algorithms are functions used to compute allocations - % of your portfolio or maximum investment size to
 invest in a market or asset.
"""

# External packages
import numpy as np
import pandas as pd
import inspect
import git
from typing import List, Callable, Dict


# InferStat packages
from infertrade.PandasEnum import PandasEnum
import infertrade.utilities.operations as operations
import infertrade.algos.community.signals as signals


def fifty_fifty(dataframe) -> pd.DataFrame:
    """Allocates 50% of strategy budget to asset, 50% to cash."""
    dataframe["allocation"] = 0.5
    return dataframe


def buy_and_hold(dataframe: pd.DataFrame) -> pd.DataFrame:
    """Allocates 100% of strategy budget to asset, holding to end of period (or security bankruptcy)."""
    dataframe[PandasEnum.ALLOCATION.value] = 1.0
    return dataframe


def chande_kroll_crossover_strategy(
    dataframe: pd.DataFrame,
) -> pd.DataFrame:
    """
    This simple all-or-nothing rule:
    (1) allocates 100% of the portofolio to a long position on the asset when the price of the asset is above both the
    Chande Kroll stop long line and Chande Kroll stop short line, and
    (2) according to the value set for the allow_short_selling parameter, either allocates 0% of the portofiolio to
    the asset or allocates 100% of the portfolio to a short position on the asset when the price of the asset is below
    both the Chande Kroll stop long line and the Chande Kroll stop short line.
    """
    # Calculate the Chande Kroll lines, which will be added to the DataFrame as columns named "chande_kroll_long" and
    # "chande_kroll_short".
    dataframe = signals.chande_kroll(dataframe)

    # Allocate positions according to the Chande Kroll lines
    is_price_above_lines = (dataframe["price"] > dataframe["chande_kroll_long"]) & (
        dataframe["price"] > dataframe["chande_kroll_short"]
    )
    is_price_below_lines = (dataframe["price"] < dataframe["chande_kroll_long"]) & (
        dataframe["price"] < dataframe["chande_kroll_short"]
    )

    dataframe.loc[is_price_above_lines, PandasEnum.ALLOCATION.value] = 1.0
    dataframe.loc[is_price_below_lines, PandasEnum.ALLOCATION.value] = -1.0

    # Delete the columns with the Chande Kroll indicators before returning
    dataframe.drop(columns=["chande_kroll_long", "chande_kroll_short"], inplace=True)

    return dataframe


def change_relationship(dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates a change relationship, which compares the asset's future price change to the last change in the signal series.

    Notes:
    - Does not fill NaNs in input, so full data needs to be supplied.
    - Error estimation uses same window as used for calibrating regression coefficients
    """
    df = dataframe.copy()
    regression_period = 120
    minimum_length_to_calculate = regression_period + 1

    if len(df[PandasEnum.MID.value]) < minimum_length_to_calculate:
        df[PandasEnum.ALLOCATION.value] = 0.0
        return df

    df = calculate_change_relationship(df, regression_period)

    return df


def calculate_change_relationship(df: pd.DataFrame, regression_period: int = 120, kelly_fraction: float = 1.0):
    """Calculates allocations for change relationship."""
    dataframe = df.copy()
    dataframe[PandasEnum.SIGNAL.value] = dataframe["research"]
    forecast_period = 100
    signal_lagged = operations.lag(
        np.reshape(dataframe[PandasEnum.SIGNAL.value].append(pd.Series([0])).values, (-1, 1)), shift=1
    )
    signal_lagged_pct_change = operations.pct_chg(signal_lagged)
    signal_lagged_pct_change[0] = [0.0]
    signal_lagged_pct_change[1] = [0.0]
    last_feature_row = signal_lagged_pct_change[-1:]
    signal_lagged_pct_change = signal_lagged_pct_change[:-1]
    price_pct_chg = operations.pct_chg(dataframe[PandasEnum.MID.value])
    price_pct_chg[0] = [0.0]

    dataframe = operations.calculate_regression_with_kelly_optimum(
        dataframe,
        feature_matrix=signal_lagged_pct_change,
        last_feature_row=last_feature_row,
        target_array=price_pct_chg,
        regression_period=regression_period,
        forecast_period=forecast_period,
        kelly_fraction=kelly_fraction,
    )
    return dataframe


def combination_relationship(dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates a combination relationship, which compares the asset's future price change to the multivariate regression of the level of the signal, the last change in the signal and the difference between the signal and the price.

    Notes:
    - Does not fill NaNs in input, so full data needs to be supplied.
    - Error estimation uses same window as used for calibrating regression coefficients
    """

    df = dataframe.copy()
    regression_period = 120
    minimum_length_to_calculate = regression_period + 1
    if len(df[PandasEnum.MID.value]) < minimum_length_to_calculate:
        df[PandasEnum.ALLOCATION.value] = 0.0
        return df

    df = calculate_combination_relationship(df, regression_period)

    return df


def calculate_combination_relationship(df: pd.DataFrame, regression_period: int = 120, kelly_fraction: float = 1.0):
    """Calculates allocations for combination relationship."""
    dataframe = df.copy()
    dataframe[PandasEnum.SIGNAL.value] = dataframe.loc[:, "research"]
    forecast_period = 100
    signal_lagged = operations.lag(
        np.reshape(dataframe[PandasEnum.SIGNAL.value].append(pd.Series([0])).values, (-1, 1)), shift=1
    )
    signal_lagged[0] = [0.0]
    signal_lagged_pct_change = operations.pct_chg(signal_lagged)
    signal_lagged_pct_change[0] = [0.0]
    signal_lagged_pct_change[1] = [0.0]
    signal_differenced = operations.research_over_price_minus_one(
        np.column_stack(
            (
                dataframe[PandasEnum.MID.value].append(pd.Series([0])).values,
                dataframe[PandasEnum.SIGNAL.value].append(pd.Series([0])).values,
            )
        ),
        shift=1,
    )
    signal_differenced[0] = [0.0]
    intermediate_matrix = np.column_stack((signal_lagged, signal_lagged_pct_change, signal_differenced))
    last_feature_row = intermediate_matrix[-1:]
    intermediate_matrix = intermediate_matrix[:-1]
    price_pct_chg = operations.pct_chg(dataframe[PandasEnum.MID.value])
    price_pct_chg[0] = [0.0]

    dataframe = operations.calculate_regression_with_kelly_optimum(
        dataframe,
        feature_matrix=intermediate_matrix,
        last_feature_row=last_feature_row,
        target_array=price_pct_chg,
        regression_period=regression_period,
        forecast_period=forecast_period,
        kelly_fraction=kelly_fraction,
    )
    return dataframe


def constant_allocation_size(dataframe: pd.DataFrame, fixed_allocation_size: float = 1.0) -> pd.DataFrame:
    """
    Returns a constant allocation, controlled by the fixed_allocation_size parameter.

    parameters:
    fixed_allocation_size: determines allocation size.
    """
    dataframe[PandasEnum.ALLOCATION.value] = fixed_allocation_size
    return dataframe


def difference_relationship(dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates a difference relationship, which compares the asset's future price change to the last difference between the signal series and asset price.

    Notes:
    - Does not fill NaNs in input, so full data needs to be supplied.
    - Error estimation uses same window as used for calibrating regression coefficients
    """

    df = dataframe.copy()
    regression_period = 120
    minimum_length_to_calculate = regression_period + 1
    if len(df[PandasEnum.MID.value]) < minimum_length_to_calculate:
        df[PandasEnum.ALLOCATION.value] = 0.0
        return df

    df = calculate_difference_relationship(df, regression_period)

    return df


def calculate_difference_relationship(df: pd.DataFrame, regression_period: int = 120, kelly_fraction: float = 1.0):
    """Calculates allocations for difference relationship."""
    dataframe = df.copy()
    dataframe[PandasEnum.SIGNAL.value] = dataframe["research"]
    forecast_period = 100
    signal_differenced = operations.research_over_price_minus_one(
        np.column_stack(
            (
                dataframe[PandasEnum.MID.value].append(pd.Series([0])).values,
                dataframe[PandasEnum.SIGNAL.value].append(pd.Series([0])).values,
            )
        ),
        shift=1,
    )
    signal_differenced[0] = [0.0]
    last_feature_row = signal_differenced[-1:]
    signal_differenced = signal_differenced[:-1]
    price_pct_chg = operations.pct_chg(dataframe[PandasEnum.MID.value])
    price_pct_chg[0] = [0.0]

    dataframe = operations.calculate_regression_with_kelly_optimum(
        dataframe,
        feature_matrix=signal_differenced,
        last_feature_row=last_feature_row,
        target_array=price_pct_chg,
        regression_period=regression_period,
        forecast_period=forecast_period,
        kelly_fraction=kelly_fraction,
    )
    return dataframe


def high_low_difference(dataframe: pd.DataFrame, scale: float = 1.0, constant: float = 0.0) -> pd.DataFrame:
    """
    Returns an allocation based on the difference in high and low values. This has been added as an
    example with multiple series and parameters.

    parameters:
    scale: determines amplitude factor.
    constant: scalar value added to the allocation size.
    """
    dataframe[PandasEnum.ALLOCATION.value] = (dataframe["high"] - dataframe["low"]) * scale + constant
    return dataframe


def level_relationship(dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates a level relationship, which compares the asset's future price change to the last value of the signal series.

    Notes:
    - Does not fill NaNs in input, so full data needs to be supplied.
    - Error estimation uses same window as used for calibrating regression coefficients
    """

    df = dataframe.copy()
    regression_period = 120
    minimum_length_to_calculate = regression_period + 1
    if len(df[PandasEnum.MID.value]) < minimum_length_to_calculate:
        df[PandasEnum.ALLOCATION.value] = 0.0
        return df

    df = calculate_level_relationship(df, regression_period)

    return df


def calculate_level_relationship(df: pd.DataFrame, regression_period: int = 120, kelly_fraction: float = 1.0):
    """Calculates allocations for level relationship."""
    dataframe = df.copy()
    dataframe[PandasEnum.SIGNAL.value] = dataframe.loc[:, "research"]
    forecast_period = 100
    signal_lagged = operations.lag(
        np.reshape(dataframe[PandasEnum.SIGNAL.value].append(pd.Series([0])).values, (-1, 1)), shift=1
    )  # revert back to manually calculating last row? doing it manually seems awkward, doing it this way seems
    # wasteful, altering the the lag (or other) function seems hacky
    signal_lagged[0] = [0.0]
    last_feature_row = signal_lagged[-1:]
    signal_lagged = signal_lagged[:-1]
    price_pct_chg = operations.pct_chg(dataframe[PandasEnum.MID.value])
    price_pct_chg[0] = [0.0]

    dataframe = operations.calculate_regression_with_kelly_optimum(
        dataframe,
        feature_matrix=signal_lagged,
        last_feature_row=last_feature_row,
        target_array=price_pct_chg,
        regression_period=regression_period,
        forecast_period=forecast_period,
        kelly_fraction=kelly_fraction,
    )
    return dataframe


def sma_crossover_strategy(dataframe: pd.DataFrame, fast: int = 0, slow: int = 0) -> pd.DataFrame:
    """
    A Simple Moving Average crossover strategy, buys when short-term SMA crosses over a long-term SMA.

    parameters:
    fast: determines the number of periods to be included in the short-term SMA.
    slow: determines the number of periods to be included in the long-term SMA.
    """

    # Set price to dataframe price column
    price = dataframe["price"]

    # Compute Fast and Slow SMA
    fast_sma = price.rolling(window=fast, min_periods=fast).mean()
    slow_sma = price.rolling(window=slow, min_periods=slow).mean()
    position = np.where(fast_sma > slow_sma, 1.0, 0.0)
    dataframe[PandasEnum.ALLOCATION.value] = position
    return dataframe


def weighted_moving_averages(
    dataframe: pd.DataFrame,
    avg_price_coeff: float = 1.0,
    avg_research_coeff: float = 1.0,
    avg_price_length: int = 2,
    avg_research_length: int = 2,
) -> pd.DataFrame:
    """
    This rule uses weightings of two moving averages to determine trade positioning.

    The parameters accepted are the integer lengths of each average (2 parameters - one for price, one for the research
    signal) and two corresponding coefficients that determine each average's weighting contribution. The total sum is
    divided by the current price to calculate a position size.

    This strategy is suitable where the dimensionality of the signal/research series is the same as the dimensionality
    of the price series, e.g. where the signal is a price forecast or fair value estimate of the market or security.

    parameters:
    avg_price_coeff: price contribution scalar multiple.
    avg_research_coeff: research or signal contribution scalar multiple.
    avg_price_length: determines length of average of price series.
    avg_research_length: determines length of average of research series.
    """

    # Splits out the price/research df to individual pandas Series.
    price = dataframe[PandasEnum.MID.value]
    research = dataframe["research"]

    # Calculates the averages.
    avg_price = price.rolling(window=avg_price_length).mean()
    avg_research = research.rolling(window=avg_research_length).mean()

    # Weights each average by the scalar coefficients.
    price_total = avg_price_coeff * avg_price
    research_total = avg_research_coeff * avg_research

    # Sums the contributions and normalises by the level of the price.
    # N.B. as summing, this approach assumes that research signal is of same dimensionality as the price.
    position = (price_total + research_total) / price.values
    dataframe[PandasEnum.ALLOCATION.value] = position
    return dataframe


def change_regression(
    dataframe: pd.DataFrame, change_coefficient: float = 0.1, change_constant: float = 0.1
) -> pd.DataFrame:
    """
    This is a regression-type approach that directly calculates allocation from change in the research level.

    parameters:
    change_coefficient: The coefficient for allocation size versus the prior day fractional change in the research.
    change_constant: The coefficient for the constant contribution.
    """
    research = dataframe["research"]
    position = (research / research.shift(1) - 1) * change_coefficient + change_constant
    dataframe[PandasEnum.ALLOCATION.value] = position
    return dataframe


def difference_regression(
    dataframe: pd.DataFrame, difference_coefficient: float = 0.1, difference_constant: float = 0.1
) -> pd.DataFrame:
    """
    This trading rules regresses the 1-day price changes seen historical against the prior day's % change
    of the research series.

    parameters:
    difference_coefficient: The coefficient for dependence on the log gap between the signal series and the price series.
    difference_constant: The coefficient for the constant contribution.
    """
    research = dataframe["research"]
    price = dataframe["price"]
    position = (research / price - 1) * difference_coefficient + difference_constant
    dataframe[PandasEnum.ALLOCATION.value] = position
    return dataframe


def level_regression(
    dataframe: pd.DataFrame, level_coefficient: float = 0.1, level_constant: float = 0.1
) -> pd.DataFrame:
    """
    This is a regression-type approach that directly calculates allocation from research level.

    parameters:
    level_coefficient: The coefficient for allocation size versus the level of the signal.
    level_constant: The coefficient for the constant contribution.
    """

    research = dataframe["research"]
    position = research * level_coefficient + level_constant
    dataframe[PandasEnum.ALLOCATION.value] = position
    return dataframe


def level_and_change_regression(
    dataframe: pd.DataFrame,
    level_coefficient: float = 0.1,
    change_coefficient: float = 0.1,
    level_and_change_constant: float = 0.1,
) -> pd.DataFrame:
    """
    This trading rules regresses the 1-day price changes seen historical against the prior day's % change of the
    research series and level of research series.

    parameters:
    level_coefficient: The coefficient for allocation size versus the level of the signal.
    change_coefficient: The coefficient for allocation size versus the prior day fractional change in the research.
    level_and_change_constant: The coefficient for the constant contribution.
    """

    research = dataframe["research"]
    position = (
        research * level_coefficient
        + (research / research.shift(1) - 1) * change_coefficient
        + level_and_change_constant
    )
    dataframe[PandasEnum.ALLOCATION.value] = position
    return dataframe


def buy_golden_cross_sell_death_cross(
    df: pd.DataFrame,
    allocation_size: float = 0.5,
    deallocation_size: float = 0.5,
    short_term_moving_avg_length: int = 50,
    long_term_moving_avg_length: int = 200,
) -> pd.DataFrame:
    """
    This trading rule allocates specified percentage of strategy budget to asset when there is a golden cross
    and deallocates specified percentage of strategy budget from asset when there is a death cross.

    Allocation and deallocation percentages specified in the parameters. Moving average lengths also
    specified in the parameters.

    parameters:
    allocation_size: The percentage of strategy budget to be allocated to asset upon golden cross
    deallocation_size: The percentage of strategy budget to deallocate from asset upon death cross
    short_term_moving_avg_length: The number of days for the short-term moving average length (default: 50 days)
    long_term_moving_avg_length: The number of days for the long-term moving average length (default: 200 days)
    """

    short_term_df = df["price"].rolling(short_term_moving_avg_length).mean()
    long_term_df = df["price"].rolling(long_term_moving_avg_length).mean()

    for i in range(long_term_moving_avg_length + 1, len(df["price"])):
        if short_term_df[i] >= long_term_df[i] and short_term_df[i - 1] < long_term_df[i - 1]:
            df.at[i, "allocation"] = allocation_size
        elif short_term_df[i] <= long_term_df[i] and short_term_df[i - 1] > long_term_df[i - 1]:
            df.at[i, "allocation"] = -deallocation_size

    return df


def SMA_strategy(df: pd.DataFrame, window: int = 1, max_investment: float = 0.1) -> pd.DataFrame:
    """
    Simple simple moving average strategy which buys when price is above signal and sells when price is below signal
    """
    sma = signals.simple_moving_average(df, window=window)["signal"]

    price_above_signal = df["close"] > sma
    price_below_signal = df["close"] <= sma

    df.loc[price_above_signal, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[price_below_signal, PandasEnum.ALLOCATION.value] = -max_investment
    return df


def WMA_strategy(df: pd.DataFrame, window: int = 1, max_investment: float = 0.1) -> pd.DataFrame:

    """
    Weighted moving average strategy which buys when price is above signal and sells when price is below signal
    """
    wma = signals.weighted_moving_average(df, window=window)["signal"]

    price_above_signal = df["close"] > wma
    price_below_signal = df["close"] <= wma

    df.loc[price_above_signal, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[price_below_signal, PandasEnum.ALLOCATION.value] = -max_investment
    return df


def MACD_strategy(
    df: pd.DataFrame, window_slow: int = 26, window_fast: int = 12, window_signal: int = 9, max_investment: float = 0.1
) -> pd.DataFrame:
    """
    Moving average convergence divergence strategy which buys when MACD signal is above 0 and sells when MACD signal
     is below zero.
    """
    macd_signal = signals.moving_average_convergence_divergence(df, window_slow, window_fast, window_signal)["signal"]

    signal_above_zero_line = macd_signal > 0
    signal_below_zero_line = macd_signal <= 0

    df.loc[signal_above_zero_line, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[signal_below_zero_line, PandasEnum.ALLOCATION.value] = -max_investment
    return df


def RSI_strategy(df: pd.DataFrame, window: int = 14, max_investment: float = 0.1) -> pd.DataFrame:
    """
    Relative Strength Index
    """
    # https://www.investopedia.com/terms/r/rsi.asp
    rsi = signals.relative_strength_index(df, window=window)["signal"]

    over_valued = rsi >= 70
    under_valued = rsi <= 30
    hold = rsi.between(30, 70)

    df.loc[over_valued, PandasEnum.ALLOCATION.value] = -max_investment
    df.loc[under_valued, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[hold, PandasEnum.ALLOCATION.value] = 0.0
    return df


def stochastic_RSI_strategy(df: pd.DataFrame, window: int = 14, max_investment: float = 0.1) -> pd.DataFrame:
    """
    Stochastic Relative Strength Index Strategy
    """
    # https://www.investopedia.com/terms/s/stochrsi.asp

    stoch_rsi = signals.stochastic_relative_strength_index(df, window=window)["signal"]

    over_valued = stoch_rsi >= 0.8
    under_valued = stoch_rsi <= 0.2
    hold = stoch_rsi.between(0.2, 0.8)

    df.loc[over_valued, PandasEnum.ALLOCATION.value] = -max_investment
    df.loc[under_valued, PandasEnum.ALLOCATION.value] = max_investment

    df.loc[hold, PandasEnum.ALLOCATION.value] = 0.0
    return df


def EMA_strategy(df: pd.DataFrame, window: int = 50, max_investment: float = 0.1) -> pd.DataFrame:
    """
    Exponential moving average strategy which buys when price is above signal and sells when price is below signal
    """
    ema = signals.exponentially_weighted_moving_average(df, window=window)["signal"]

    price_above_signal = df["close"] > ema
    price_below_signal = df["close"] <= ema

    df.loc[price_above_signal, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[price_below_signal, PandasEnum.ALLOCATION.value] = -max_investment
    return df


def bollinger_band_strategy(
    df: pd.DataFrame, window: int = 20, window_dev: int = 2, max_investment: float = 0.1
) -> pd.DataFrame:

    """
    This is Strategy that identify overbought or oversold market conditions.
        1. Oversold: Price breaks below the lower band of the Bollinger Bands
        2. Overbought: Price breaks above the upper band of the Bollinger bands

    Relies on concept "Mean reversion"
    Reference: https://www.investopedia.com/trading/using-bollinger-bands-to-gauge-trends/
    """
    short_position = False
    long_position = False
    df_with_signal = signals.bollinger_band(df, window=window, window_dev=window_dev)
    for index, row in df_with_signal.iterrows():

        # Check for short position
        if (row["typical_price"] >= row["BOLU"] or short_position == True) and row["typical_price"] > row["BOLA"]:
            short_position = True
        else:
            short_position = False

        # Check for long position
        if (row["typical_price"] <= row["BOLD"] or long_position == True) and row["typical_price"] < row["BOLA"]:
            long_position = True
        else:
            long_position = False

        # Both short position and long position can't be true
        assert not (short_position and long_position)

        # allocation conditions
        if short_position:
            df.loc[index, PandasEnum.ALLOCATION.value] = max_investment

        elif long_position:
            df.loc[index, PandasEnum.ALLOCATION.value] = -max_investment

        else:
            # if both short position and long position is false
            df.loc[index, PandasEnum.ALLOCATION.value] = 0.0

    return df


def DPO_strategy(df: pd.DataFrame, window: int = 20, max_investment: float = 0.1) -> pd.DataFrame:
    """
    Exponential moving average strategy which buys when price is above signal and sells when price is below signal.
    """
    dpo = signals.detrended_price_oscillator(df, window=window)["signal"]

    above_zero = dpo > 0
    below_zero = dpo <= 0

    df.loc[above_zero, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[below_zero, PandasEnum.ALLOCATION.value] = -max_investment
    return df


def PPO_strategy(
    df: pd.DataFrame, window_slow: int = 26, window_fast: int = 12, window_signal: int = 9, max_investment: float = 0.1
) -> pd.DataFrame:
    """
    Percentage Price Oscillator strategy which buys when signal is above zero and sells when signal is below zero
    """
    ppo = signals.percentage_price_oscillator(df, window_slow, window_fast, window_signal)["signal"]

    above_zero = ppo > 0
    below_zero = ppo <= 0

    df.loc[above_zero, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[below_zero, PandasEnum.ALLOCATION.value] = -max_investment
    return df


def PVO_strategy(
    df: pd.DataFrame, window_slow: int = 26, window_fast: int = 12, window_signal: int = 9, max_investment: float = 0.1
) -> pd.DataFrame:
    """
    Percentage volume Oscillator strategy which buys when signal is above zero and sells when signal is below zero
    """
    PVO = signals.percentage_volume_oscillator(df, window_slow, window_fast, window_signal)["signal"]

    above_zero = PVO > 0
    below_zero = PVO <= 0

    df.loc[above_zero, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[below_zero, PandasEnum.ALLOCATION.value] = -max_investment
    return df


def TRIX_strategy(df: pd.DataFrame, window: int = 14, max_investment: float = 0.1) -> pd.DataFrame:
    """
    This is Triple Exponential Average (TRIX) strategy which buys when signal is above zero and sells when signal is below zero
    """
    trix = signals.triple_exponential_average(df, window)["signal"]

    above_zero = trix > 0
    below_zero = trix <= 0

    df.loc[above_zero, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[below_zero, PandasEnum.ALLOCATION.value] = -max_investment
    return df


def TSI_strategy(
    df: pd.DataFrame, window_slow: int = 25, window_fast: int = 13, window_signal: int = 13, max_investment: float = 0.1
) -> pd.DataFrame:
    """
    This is True Strength Index (TSI) strategy which buys when TSI is greater than signal and sells when TSI is below signal
    Signal is EMA of TSI
    """
    df_with_signals = signals.true_strength_index(df, window_slow, window_fast, window_signal)

    above_signal = df_with_signals["TSI"] > df_with_signals["signal"]
    below_signal = df_with_signals["TSI"] <= df_with_signals["signal"]

    df.loc[above_signal, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[below_signal, PandasEnum.ALLOCATION.value] = -max_investment
    return df


def STC_strategy(
    df: pd.DataFrame,
    window_slow: int = 50,
    window_fast: int = 23,
    cycle: int = 10,
    smooth1: int = 3,
    smooth2: int = 3,
    max_investment: float = 0.1,
) -> pd.DataFrame:
    """
    This is Schaff Trend Cycle (STC) strategy which indicate
        1. oversold when STC < 25
        2. overbought when STC > 75
    """
    stc = signals.schaff_trend_cycle(df, window_slow, window_fast, cycle, smooth1, smooth2)["signal"]

    oversold = stc <= 25
    overbought = stc >= 75
    hold = stc.between(25, 75)

    df.loc[oversold, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[overbought, PandasEnum.ALLOCATION.value] = -max_investment
    df.loc[hold, PandasEnum.ALLOCATION.value] = 0

    return df


def KAMA_strategy(
    df: pd.DataFrame, window: int = 10, pow1: int = 2, pow2: int = 30, max_investment: float = 0.1
) -> pd.DataFrame:
    """
    Kaufman's Adaptive Moving Average (KAMA) strategy indicates
        1. downtrend when signal < price
        2. uptrend when signal > price
    """
    df_with_signals = signals.KAMA(df, window, pow1, pow2)

    downtrend = df_with_signals["signal"] <= df_with_signals["close"]
    uptrend = df_with_signals["signal"] > df_with_signals["close"]

    df.loc[uptrend, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[downtrend, PandasEnum.ALLOCATION.value] = -max_investment

    return df


def aroon_strategy(df: pd.DataFrame, window: int = 25, max_investment: float = 0.1) -> pd.DataFrame:
    """
    The Arron indicator is composed of two lines.
        1. Aroon_up: line which measures the number of periods since a High, and
        2. Aroon_down: line which measures the number of periods since a Low.

    This strategy indicates:
        1. Bearish: when aroon_up < aroon_down
        2. Bullish: when aroon_up >= aroon_down
    """
    df_with_signals = signals.aroon(df, window)

    bullish = df_with_signals["aroon_up"] >= df_with_signals["aroon_down"]
    bearish = df_with_signals["aroon_down"] < df_with_signals["aroon_up"]

    df.loc[bullish, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[bearish, PandasEnum.ALLOCATION.value] = -max_investment

    return df


def ROC_strategy(df: pd.DataFrame, window: int = 12, max_investment: float = 0.1) -> pd.DataFrame:
    """
    A rising ROC above zero typically confirms an uptrend while a falling ROC below zero indicates a downtrend.
    """
    df_with_signals = signals.rate_of_change(df, window)

    uptrend = df_with_signals["signal"] >= 0
    downtrend = df_with_signals["signal"] < 0

    df.loc[uptrend, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[downtrend, PandasEnum.ALLOCATION.value] = -max_investment

    return df


def ADX_strategy(df: pd.DataFrame, window: int = 14, max_investment: float = 0.1) -> pd.DataFrame:
    """
    Average Directional Movement Index makes use of three indicators to measure both trend direction and its strength.
        1. Plus Directional Indicator (+DI)
        2. Negative Directonal Indicator (-DI)
        3. Average directional Index (ADX)

    +DI and -DI measures the trend direction and ADX measures the strength of trend
    """
    df_with_signals = signals.average_directional_movement_index(df, window)

    plus_di = df_with_signals["ADX_POS"]
    minus_di = df_with_signals["ADX_NEG"]
    adx = df_with_signals["ADX"]

    index = 0
    for pdi_value, mdi_value, adx_value in zip(plus_di, minus_di, adx):
        # ADX > 25 to avoid risky investment i.e. invest only when trend is strong
        if adx_value > 25 and pdi_value > mdi_value:
            df.loc[index, PandasEnum.ALLOCATION.value] = max_investment

        elif adx_value > 25 and pdi_value < mdi_value:
            df.loc[index, PandasEnum.ALLOCATION.value] = -max_investment

        else:
            df.loc[index, PandasEnum.ALLOCATION.value] = 0

        index += 1

    return df


def vortex_strategy(df: pd.DataFrame, window: int = 14, max_investment: float = 0.1) -> pd.DataFrame:
    """
    A rising ROC above zero typically confirms an uptrend while a falling ROC below zero indicates a downtrend.
    """
    df_with_signals = signals.vortex_indicator(df, window)

    uptrend = df_with_signals["VORTEX_POS"] >= df_with_signals["VORTEX_NEG"]
    downtrend = df_with_signals["VORTEX_POS"] < df_with_signals["VORTEX_NEG"]

    df.loc[uptrend, PandasEnum.ALLOCATION.value] = max_investment
    df.loc[downtrend, PandasEnum.ALLOCATION.value] = -max_investment

    return df


# Below we populate a function list and a required series list. This needs to be updated when adding new rules.

function_list = [
    fifty_fifty,
    buy_and_hold,
    chande_kroll_crossover_strategy,
    change_relationship,
    combination_relationship,
    difference_relationship,
    level_relationship,
    constant_allocation_size,
    high_low_difference,
    sma_crossover_strategy,
    weighted_moving_averages,
    change_regression,
    difference_regression,
    level_regression,
    level_and_change_regression,
    buy_golden_cross_sell_death_cross,
    SMA_strategy,
    WMA_strategy,
    MACD_strategy,
    RSI_strategy,
    stochastic_RSI_strategy,
    EMA_strategy,
    bollinger_band_strategy,
    PPO_strategy,
    PVO_strategy,
    TRIX_strategy,
    TSI_strategy,
    STC_strategy,
    KAMA_strategy,
    aroon_strategy,
    ROC_strategy,
    ADX_strategy,
    vortex_strategy,
    DPO_strategy,
]

required_series_dict = {
        "fifty_fifty": [],
        "buy_and_hold": [],
        "chande_kroll_crossover_strategy": ["high", "low", "price"],
        "change_relationship": ["price", "research"],
        "combination_relationship": ["price", "research"],
        "difference_relationship": ["price", "research"],
        "level_relationship": ["price", "research"],
        "constant_allocation_size": [],
        "high_low_difference": ["high", "low"],
        "sma_crossover_strategy": ["price"],
        "weighted_moving_averages": ["price", "research"],
        "change_regression": ["research"],
        "difference_regression": ["price", "research"],
        "level_regression": ["research"],
        "level_and_change_regression": ["research"],
        "buy_golden_cross_sell_death_cross": ["price"],
        "SMA_strategy": ["close"],
        "WMA_strategy": ["close"],
        "MACD_strategy": ["close"],
        "RSI_strategy": ["close"],
        "stochastic_RSI_strategy": ["close"],
        "EMA_strategy": ["close"],
        "bollinger_band_strategy": ["close"],
        "PPO_strategy": ["close"],
        "PVO_strategy": ["volume"],
        "TRIX_strategy": ["close"],
        "TSI_strategy": ["close"],
        "STC_strategy": ["close"],
        "KAMA_strategy": ["close"],
        "aroon_strategy": ["close"],
        "ROC_strategy": ["close"],
        "ADX_strategy": ["close", "high", "low"],
        "vortex_strategy": ["close", "high", "low"],
        "DPO_strategy": ["close"],
    }


def get_functions_list() -> List[Callable]:
    """Returns list of functions."""
    return function_list


def get_required_series() -> Dict[str, list]:
    """Returns dictionary of series"""
    return required_series_dict


def get_functions_names() -> List[str]:
    """Returns list of functions"""
    series_dict = get_required_series()
    list_of_functions = list(series_dict.keys())
    return list_of_functions


def get_latest_infertrade_commit() -> str:
    """Gets the latest commit for InferTrade as a string."""
    repo = git.Repo(search_parent_directories=True)
    commit = str(repo.head.commit)
    return commit


def get_latest_infertrade_allocation_file_url() -> str:
    """Gets the latest URL stub for the allocation file."""
    github_permalink = (
        "https://github.com/ta-oliver/infertrade/blob/"
        + get_latest_infertrade_commit()
        + "/infertrade/algos/community/allocations.py"
    )
    return github_permalink


def create_permalink_to_allocations(function: Callable) -> str:
    """Creates a permalink to the referenced function."""
    full_link = get_latest_infertrade_allocation_file_url() + "#L" + str(function.__code__.co_firstlineno)
    return full_link


def get_parameters(function: Callable) -> dict:
    """Gets the default parameters and its values from the function"""
    signature = inspect.signature(function)
    parameter_items = signature.parameters.items()
    is_empty = inspect.Parameter.empty
    parameters = {key: value.default for key, value in parameter_items if value.default is not is_empty}
    return parameters


def create_infertrade_export_allocations():
    """Creates a dictionary for export."""
    infertrade_export_allocations_raw = {}
    list_of_functions = get_functions_list()
    series_dict = get_required_series()
    for function in list_of_functions:

        infertrade_export_allocations_raw.update(
            {
                function.__name__: {
                    "function": function,
                    "parameters": get_parameters(function),
                    "series": series_dict[function.__name__],
                    "available_representation_types": {"github_permalink": create_permalink_to_allocations(function)},
                }
            }
        )

    return infertrade_export_allocations_raw


infertrade_export_allocations = create_infertrade_export_allocations()


if __name__ == "__main__":
    """To quickly view rule properties."""
    print(infertrade_export_allocations)



#!/usr/bin/env python3
"""
infertrade setup.py file

   Copyright 12th March 2021 InferStat Ltd

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

# Python standard library
import re

from os import walk
from os.path import dirname, abspath, join
from pathlib import Path

# Third library (not InferStat)
from setuptools import find_packages, setup

# Hardcoded variables
PROJECT_NAME = "infertrade"
PROJECT_DESCRIPTION = "Pandas and SciKit Learn compatible open source interface for algorithmic trading functions."
BLACKLIST_DIRS = ["example_scripts"]

this_directory = Path(dirname(abspath(__file__)))


def get_long_description(filename: str = "README.md") -> str:
    """Returns a long repository description read from file."""
    try:
        with open(Path.joinpath(this_directory, filename)) as f:
            long_description = "\n" + f.read()
    except FileNotFoundError:
        long_description = ""
    return long_description


def get_pkg_list(requirement_file: str) -> list:
    """Returns a list of packages requirements from listed file."""
    with open(Path.joinpath(this_directory, requirement_file)) as f:
        packages_full_list = [pkg.strip() for pkg in f.readlines() if not pkg.startswith("#")]
        list_of_packages_with_versions = [pkg for pkg in packages_full_list if pkg]
    return list_of_packages_with_versions


def get_version(filename: str = "_version.py") -> str:
    """Returns the package version number as a string by searching and reading the _version.py file."""
    gitignore = []
    for dirpath, _, filenames in walk(this_directory):
        if ".gitignore" in filenames:
            with open(join(this_directory,".gitignore")) as _f:
                gitignore = [file.strip() for file in _f.readlines() if not re.search(r"\#|\*", file)]

        if any(pattern for pattern in gitignore + BLACKLIST_DIRS if re.search(pattern, dirpath)):
            continue
        for file in filenames:
            if filename in file:
                file_path = Path().joinpath(dirpath, filename)
    try:
        assert file_path.is_file()
        with open(file_path) as f:
            _version_info = "".join([i.strip() for i in f.readlines() if i.startswith("_")])
            _version_info = _version_info.replace(" ", "").replace('"', "").replace('"', "")
            about = dict([_version_info.split("=")])  # noqa: C406
    except FileNotFoundError:  # TODO - probably should be expected error types.
        raise RuntimeError(f"Unable to find version information in '{file_path}'.")
    else:  # TODO - currently this can't trigger as prior exception catches all errors.
        return about["__version__"]


# Get a list of packages as defined in file
package_requirements = get_pkg_list("requirements.txt")
dev_requirements = get_pkg_list("requirements-dev.txt")

# Setting up basic parameters of infertrade library
setup(
    name=PROJECT_NAME,
    version=get_version(),
    description=PROJECT_DESCRIPTION,
    license="Apache License 2.0",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url=f"https://github.com/ta-oliver/infertrade",
    author="Thomas Oliver",
    author_email="support@infertrade.com",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=package_requirements,
    extras_require={"dev": dev_requirements},
    tests_require=["pytest"],
    python_requires=">=3.7.0,<3.8.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Natural Language :: English",
    ],
    zip_safe=False,
)
